; CONSOLE CLEAN 
; 

    include     ASMRT.inc
    includelib  ASMRT.lib
    
    include     ptimers.inc
    
; 
    .data?
        hHeap   DWORD   ?
        hLFHeap DWORD   ?
        
        spHeader    DWORD   ?
        spHeap      DWORD   ?
        spLFHeap    DWORD   ?
        spHeapOvh   DWORD   ?
        spLFHeapOvh DWORD   ?

; 

PHE_BLOCK STRUCT
    hMem        DWORD   ?
    dwReserved  DWORD   3 DUP (?)
PHE_BLOCK ENDS

PHE_REGION STRUCT
    dwCommittedSize     DWORD   ?
    dwUnCommittedSize   DWORD   ?
    lpFirstBlock        DWORD   ?
    lpLastBlock         DWORD   ?
PHE_REGION ENDS

PROCESS_HEAP_ENTRY STRUCT
    lpData          DWORD   ?
    cbData          DWORD   ?
    cbOverhead      BYTE    ?
    iRegionIndex    BYTE    ?
    wFlags          WORD    ?
    UNION
        Block       PHE_BLOCK   <?>
        Region      PHE_REGION  <?>
    ENDS
PROCESS_HEAP_ENTRY ENDS
        
; 

CreateHeap      PROTO
CreateLFHeap    PROTO
Measure         PROTO dwAllocSize:DWORD

; 
    .code
; 

Main PROC
    ConsoleInit
    
    ; Create a growable heap starting from 4KB with exceptions turned on
    invoke  CreateHeap
    mov     hHeap, eax
    
    ; Create a growable low-fragmentation starting from 4KB with exceptions turned on
    invoke  CreateLFHeap
    .if(!eax)
        Print   "Low fragmentation heaps are not supported.\r\nValues displayed will be for a normal heap.\r\n\r\n"
    .endif
    mov     hLFHeap, eax
    
    mov     spHeader,    Copy("               |")
    mov     spHeap,      Copy("Normal Heap    |")
    mov     spHeapOvh,   Copy("    Overhead   |")
    mov     spLFHeap,    Copy("Low Frag Heap  |")
    mov     spLFHeapOvh, Copy("    Overhead   |")
    
    mov     ecx, 7
    mov     edx, 1
    .repeat
        invoke  Measure, edx
        add     edx, edx
        dec     ecx
    .until(ZERO?)
    
    ;Print   "%[s]\r\n%[s]\r\n%[s]\r\n\r\n", spHeader, spHeap, spLFHeap
    CPrintLn    spHeader
    CPrintLn    spHeap
    CPrintLn    spHeapOvh
    CPrintLn    spLFHeap
    CPrintLn    spLFHeapOvh
    CPrintLn
    
    StringFree spHeader, spHeap, spHeapOvh, spLFHeap, spLFHeapOvh
    
    mov     spHeader,    Copy("               |")
    mov     spHeap,      Copy("Normal Heap    |")
    mov     spHeapOvh,   Copy("    Overhead   |")
    mov     spLFHeap,    Copy("Low Frag Heap  |")
    mov     spLFHeapOvh, Copy("    Overhead   |")
        
    mov     ecx, 7
    mov     edx, 1 SHL 7
    .repeat
        invoke  Measure, edx
        add     edx, edx
        dec     ecx
    .until(ZERO?)
    
    ;Print   "%[s]\r\n%[s]\r\n%[s]\r\n\r\n", spHeader, spHeap, spLFHeap
    CPrintLn    spHeader
    CPrintLn    spHeap
    CPrintLn    spHeapOvh
    CPrintLn    spLFHeap
    CPrintLn    spLFHeapOvh
    CPrintLn
    
    StringFree spHeader, spHeap, spHeapOvh, spLFHeap, spLFHeapOvh
    
    invoke  HeapDestroy, hHeap
    invoke  HeapDestroy, hLFHeap
    CPause  "Press any key to exit . . ."
    ret
Main ENDP

; 

Measure PROC USES ecx edx dwAllocSize:DWORD
    LOCAL   Entry:PROCESS_HEAP_ENTRY
    
    Format "%[s]%|8u|", spHeader, dwAllocSize
    StringFree spHeader
    mov     spHeader, eax
    
; ====================================
    PTIMER_START_LOOP_COUNT
    invoke  HeapAlloc, hHeap, 0, dwAllocSize
    invoke  HeapFree, hHeap, 0, eax
    PTIMER_END_LOOP_COUNT
    
    Format "%[s]%7q |", spHeap, qwPTIMER_RESULT
    StringFree spHeap
    mov     spHeap, eax
    
; -------------------------------------------------------------------------
    push    edi
    invoke  HeapAlloc, hHeap, 0, dwAllocSize
    mov     edi, eax
    mov     Entry.lpData, 0
    .while(Entry.lpData != edi && eax)
        lea     eax, Entry
        invoke  HeapWalk, hHeap, eax
    .endw
    
    movzx   eax, Entry.cbOverhead
    Format "%[s]%7d |", spHeapOvh, eax
    StringFree spHeapOvh
    mov     spHeapOvh, eax
    
    invoke  HeapFree, hHeap, 0, edi
    pop     edi
    
; ====================================
    PTIMER_START_LOOP_COUNT
    invoke  HeapAlloc, hLFHeap, 0, dwAllocSize
    invoke  HeapFree, hLFHeap, 0, eax
    PTIMER_END_LOOP_COUNT
    
    Format "%[s]%7q |", spLFHeap, qwPTIMER_RESULT
    StringFree spLFHeap
    mov     spLFHeap, eax
    
; -------------------------------------------------------------------------
    push    edi
    invoke  HeapAlloc, hLFHeap, 0, dwAllocSize
    mov     edi, eax
    mov     Entry.lpData, 0
    .while(Entry.lpData != edi && eax)
        lea     eax, Entry
        invoke  HeapWalk, hLFHeap, eax
    .endw
    
    movzx   eax, Entry.cbOverhead
    Format "%[s]%7d |", spLFHeapOvh, eax
    StringFree spLFHeapOvh
    mov     spLFHeapOvh, eax
    
    invoke  HeapFree, hLFHeap, 0, edi
    pop     edi
    
    ret
Measure ENDP

; 

CreateHeap PROC
    invoke  HeapCreate, 4, 4096, 0
    ret
CreateHeap ENDP

; -------------------------------------------------------------------------

CreateLFHeap PROC
    invoke  HeapCreate, 4, 4096, 0
    push    eax
    push    2
    mov     ecx, esp
    push    4
    push    ecx
    push    0
    push    eax
    PushString "HeapSetInformation"
    PushString "kernel32.dll"
    call    GetModuleHandle
    push    eax
    call    GetProcAddress
    .if(eax)
        call    eax
        pop     eax
        pop     eax
    .else
        add     esp, 24
        xor     eax, eax
    .endif
    ret
CreateLFHeap ENDP

; 
END
